@@ -4,15 +4,17 @@ import (
 	"context"
 	"errors"
 	"fmt"
-	"log"
 	"math/rand"
+	"strconv"
+	"strings"
 	"sync"
 	"sync/atomic"
 	"time"
 
-	"gopkg.in/tomb.v2"
+	tomb "gopkg.in/tomb.v2"
 
 	metrics "github.com/armon/go-metrics"
+	log "github.com/hashicorp/go-hclog"
 	multierror "github.com/hashicorp/go-multierror"
 	"github.com/hashicorp/nomad/nomad/structs"
 	"github.com/hashicorp/nomad/nomad/structs/config"
@@ -119,6 +121,9 @@ type VaultClient interface {
 	// RevokeTokens takes a set of tokens accessor and revokes the tokens
 	RevokeTokens(ctx context.Context, accessors []*structs.VaultAccessor, committed bool) error
 
+	// MarkForRevocation revokes the tokens in background
+	MarkForRevocation(accessors []*structs.VaultAccessor) error
+
 	// Stop is used to stop token renewal
 	Stop()
 
@@ -126,11 +131,11 @@ type VaultClient interface {
 	Running() bool
 
 	// Stats returns the Vault clients statistics
-	Stats() *VaultStats
+	Stats() map[string]string
 
 	// EmitStats emits that clients statistics at the given period until stopCh
 	// is called.
-	EmitStats(period time.Duration, stopCh chan struct{})
+	EmitStats(period time.Duration, stopCh <-chan struct{})
 }
 
 // VaultStats returns all the stats about Vault tokens created and managed by
@@ -139,9 +144,15 @@ type VaultStats struct {
 	// TrackedForRevoke is the count of tokens that are being tracked to be
 	// revoked since they could not be immediately revoked.
 	TrackedForRevoke int
+
+	// TokenTTL is the time-to-live duration for the current token
+	TokenTTL time.Duration
+
+	// TokenExpiry is the recorded expiry time of the current token
+	TokenExpiry time.Time
 }
 
-// PurgeVaultAccessor is called to remove VaultAccessors from the system. If
+// PurgeVaultAccessorFn is called to remove VaultAccessors from the system. If
 // the function returns an error, the token will still be tracked and revocation
 // will retry till there is a success
 type PurgeVaultAccessorFn func(accessors []*structs.VaultAccessor) error
@@ -149,12 +160,13 @@ type PurgeVaultAccessorFn func(accessors []*structs.VaultAccessor) error
 // tokenData holds the relevant information about the Vault token passed to the
 // client.
 type tokenData struct {
-	CreationTTL int      `mapstructure:"creation_ttl"`
-	TTL         int      `mapstructure:"ttl"`
-	Renewable   bool     `mapstructure:"renewable"`
-	Policies    []string `mapstructure:"policies"`
-	Role        string   `mapstructure:"role"`
-	Root        bool
+	CreationTTL   int      `mapstructure:"creation_ttl"`
+	TTL           int      `mapstructure:"ttl"`
+	Renewable     bool     `mapstructure:"renewable"`
+	Policies      []string `mapstructure:"policies"`
+	Role          string   `mapstructure:"role"`
+	NamespacePath string   `mapstructure:"namespace_path"`
+	Root          bool
 }
 
 // vaultClient is the Servers implementation of the VaultClient interface. The
@@ -165,9 +177,18 @@ type vaultClient struct {
 	// limiter is used to rate limit requests to Vault
 	limiter *rate.Limiter
 
-	// client is the Vault API client
+	// client is the Vault API client used for Namespace-relative integrations
+	// with the Vault API (anything except `/v1/sys`). If this server is not
+	// configured to reference a Vault namespace, this will point to the same
+	// client as clientSys
 	client *vapi.Client
 
+	// clientSys is the Vault API client used for non-Namespace-relative integrations
+	// with the Vault API (anything involving `/v1/sys`). This client is never configured
+	// with a Vault namespace, because these endpoints may return errors if a namespace
+	// header is provided
+	clientSys *vapi.Client
+
 	// auth is the Vault token auth API client
 	auth *vapi.TokenAuth
 
@@ -199,27 +220,42 @@ type vaultClient struct {
 	// running indicates whether the vault client is started.
 	running bool
 
+	// renewLoopActive indicates whether the renewal goroutine is running
+	// It should be accessed and updated atomically
+	// used for testing purposes only
+	renewLoopActive int32
+
 	// childTTL is the TTL for child tokens.
 	childTTL string
 
-	// lastRenewed is the time the token was last renewed
-	lastRenewed time.Time
+	// currentExpiration is the time the current token lease expires
+	currentExpiration     time.Time
+	currentExpirationLock sync.Mutex
 
 	tomb   *tomb.Tomb
-	logger *log.Logger
-
-	// stats stores the stats
-	stats     *VaultStats
-	statsLock sync.RWMutex
+	logger log.Logger
 
 	// l is used to lock the configuration aspects of the client such that
 	// multiple callers can't cause conflicting config updates
 	l sync.Mutex
+
+	// setConfigLock serializes access to the SetConfig method
+	setConfigLock sync.Mutex
+
+	// consts as struct fields for overriding in tests
+	maxRevokeBatchSize int
+	revocationIntv     time.Duration
+
+	entHandler taskClientHandler
+}
+
+type taskClientHandler interface {
+	clientForTask(v *vaultClient, namespace string) (*vapi.Client, error)
 }
 
 // NewVaultClient returns a Vault client from the given config. If the client
 // couldn't be made an error is returned.
-func NewVaultClient(c *config.VaultConfig, logger *log.Logger, purgeFn PurgeVaultAccessorFn) (*vaultClient, error) {
+func NewVaultClient(c *config.VaultConfig, logger log.Logger, purgeFn PurgeVaultAccessorFn, delegate taskClientHandler) (*vaultClient, error) {
 	if c == nil {
 		return nil, fmt.Errorf("must pass valid VaultConfig")
 	}
@@ -227,15 +263,23 @@ func NewVaultClient(c *config.VaultConfig, logger *log.Logger, purgeFn PurgeVaul
 	if logger == nil {
 		return nil, fmt.Errorf("must pass valid logger")
 	}
+	if purgeFn == nil {
+		purgeFn = func(accessors []*structs.VaultAccessor) error { return nil }
+	}
+	if delegate == nil {
+		delegate = &VaultNoopDelegate{}
+	}
 
 	v := &vaultClient{
-		config:   c,
-		logger:   logger,
-		limiter:  rate.NewLimiter(requestRateLimit, int(requestRateLimit)),
-		revoking: make(map[*structs.VaultAccessor]time.Time),
-		purgeFn:  purgeFn,
-		tomb:     &tomb.Tomb{},
-		stats:    new(VaultStats),
+		config:             c,
+		logger:             logger.Named("vault"),
+		limiter:            rate.NewLimiter(requestRateLimit, int(requestRateLimit)),
+		revoking:           make(map[*structs.VaultAccessor]time.Time),
+		purgeFn:            purgeFn,
+		tomb:               &tomb.Tomb{},
+		maxRevokeBatchSize: maxVaultRevokeBatchSize,
+		revocationIntv:     vaultRevocationIntv,
+		entHandler:         delegate,
 	}
 
 	if v.config.IsEnabled() {
@@ -294,8 +338,11 @@ func (v *vaultClient) SetActive(active bool) {
 func (v *vaultClient) flush() {
 	v.l.Lock()
 	defer v.l.Unlock()
+	v.revLock.Lock()
+	defer v.revLock.Unlock()
 
 	v.client = nil
+	v.clientSys = nil
 	v.auth = nil
 	v.connEstablished = false
 	v.connEstablishedErr = nil
@@ -312,6 +359,8 @@ func (v *vaultClient) SetConfig(config *config.VaultConfig) error {
 	if config == nil {
 		return fmt.Errorf("must pass valid VaultConfig")
 	}
+	v.setConfigLock.Lock()
+	defer v.setConfigLock.Unlock()
 
 	v.l.Lock()
 	defer v.l.Unlock()
@@ -323,12 +372,17 @@ func (v *vaultClient) SetConfig(config *config.VaultConfig) error {
 
 	// Kill any background routines
 	if v.running {
-		// Stop accepting any new request
-		v.connEstablished = false
-
-		// Kill any background routine and create a new tomb
+		// Kill any background routine
 		v.tomb.Kill(nil)
+
+		// Locking around tomb.Wait can deadlock with
+		// establishConnection exiting, so we must unlock here.
+		v.l.Unlock()
 		v.tomb.Wait()
+		v.l.Lock()
+
+		// Stop accepting any new requests
+		v.connEstablished = false
 		v.tomb = &tomb.Tomb{}
 		v.running = false
 	}
@@ -387,15 +441,29 @@ func (v *vaultClient) buildClient() error {
 	// Create the Vault API client
 	client, err := vapi.NewClient(apiConf)
 	if err != nil {
-		v.logger.Printf("[ERR] vault: failed to create Vault client. Not retrying: %v", err)
+		v.logger.Error("failed to create Vault client and not retrying", "error", err)
 		return err
 	}
 
-	// Set the token and store the client
+	// Store the client, create/assign the /sys client
+	v.client = client
+	if v.config.Namespace != "" {
+		v.logger.Debug("configuring Vault namespace", "namespace", v.config.Namespace)
+		v.clientSys, err = vapi.NewClient(apiConf)
+		if err != nil {
+			v.logger.Error("failed to create Vault sys client and not retrying", "error", err)
+			return err
+		}
+		client.SetNamespace(v.config.Namespace)
+	} else {
+		v.clientSys = client
+	}
+
+	// Set the token
 	v.token = v.config.Token
 	client.SetToken(v.token)
-	v.client = client
 	v.auth = client.Auth().Token()
+
 	return nil
 }
 
@@ -414,26 +482,27 @@ OUTER:
 		case <-v.tomb.Dying():
 			return
 		case <-retryTimer.C:
-			// Ensure the API is reachable
-			if !initStatus {
-				if _, err := v.client.Sys().InitStatus(); err != nil {
-					v.logger.Printf("[WARN] vault: failed to contact Vault API. Retrying in %v: %v",
-						v.config.ConnectionRetryIntv, err)
-					retryTimer.Reset(v.config.ConnectionRetryIntv)
-					continue OUTER
-				}
-				initStatus = true
-			}
 			// Retry validating the token till success
 			if err := v.parseSelfToken(); err != nil {
-				v.logger.Printf("[ERR] vault: failed to validate self token/role. Retrying in %v: %v", v.config.ConnectionRetryIntv, err)
+				// if parsing token fails, try to distinguish legitimate token error from transient Vault initialization/connection issue
+				if !initStatus {
+					if _, err := v.clientSys.Sys().Health(); err != nil {
+						v.logger.Warn("failed to contact Vault API", "retry", v.config.ConnectionRetryIntv, "error", err)
+						retryTimer.Reset(v.config.ConnectionRetryIntv)
+						continue OUTER
+					}
+					initStatus = true
+				}
+
+				v.logger.Error("failed to validate self token/role", "retry", v.config.ConnectionRetryIntv, "error", err)
 				retryTimer.Reset(v.config.ConnectionRetryIntv)
 				v.l.Lock()
 				v.connEstablished = true
-				v.connEstablishedErr = fmt.Errorf("Nomad Server failed to establish connections to Vault: %v", err)
+				v.connEstablishedErr = fmt.Errorf("failed to establish connection to Vault: %v", err)
 				v.l.Unlock()
 				continue OUTER
 			}
+
 			break OUTER
 		}
 	}
@@ -444,10 +513,9 @@ OUTER:
 
 	// If we are given a non-root token, start renewing it
 	if v.tokenData.Root && v.tokenData.CreationTTL == 0 {
-		v.logger.Printf("[DEBUG] vault: not renewing token as it is root")
+		v.logger.Debug("not renewing token as it is root")
 	} else {
-		v.logger.Printf("[DEBUG] vault: token lease duration is %v",
-			time.Duration(v.tokenData.CreationTTL)*time.Second)
+		v.logger.Debug("starting renewal loop", "creation_ttl", time.Duration(v.tokenData.CreationTTL)*time.Second)
 		v.tomb.Go(wrapNilError(v.renewalLoop))
 	}
 
@@ -457,9 +525,16 @@ OUTER:
 	v.l.Unlock()
 }
 
+func (v *vaultClient) isRenewLoopActive() bool {
+	return atomic.LoadInt32(&v.renewLoopActive) == 1
+}
+
 // renewalLoop runs the renew loop. This should only be called if we are given a
 // non-root token.
 func (v *vaultClient) renewalLoop() {
+	atomic.StoreInt32(&v.renewLoopActive, 1)
+	defer atomic.StoreInt32(&v.renewLoopActive, 0)
+
 	// Create the renewal timer and set initial duration to zero so it fires
 	// immediately
 	authRenewTimer := time.NewTimer(0)
@@ -474,17 +549,17 @@ func (v *vaultClient) renewalLoop() {
 			return
 		case <-authRenewTimer.C:
 			// Renew the token and determine the new expiration
-			err := v.renew()
-			currentExpiration := v.lastRenewed.Add(time.Duration(v.tokenData.CreationTTL) * time.Second)
+			recoverable, err := v.renew()
+			v.currentExpirationLock.Lock()
+			currentExpiration := v.currentExpiration
+			v.currentExpirationLock.Unlock()
 
 			// Successfully renewed
 			if err == nil {
-				// If we take the expiration (lastRenewed + auth duration) and
-				// subtract the current time, we get a duration until expiry.
-				// Set the timer to poke us after half of that time is up.
+				// Attempt to renew the token at half the expiration time
 				durationUntilRenew := currentExpiration.Sub(time.Now()) / 2
 
-				v.logger.Printf("[INFO] vault: renewing token in %v", durationUntilRenew)
+				v.logger.Info("successfully renewed token", "next_renewal", durationUntilRenew)
 				authRenewTimer.Reset(durationUntilRenew)
 
 				// Reset any backoff
@@ -492,78 +567,110 @@ func (v *vaultClient) renewalLoop() {
 				break
 			}
 
-			// Back off, increasing the amount of backoff each time. There are some rules:
-			//
-			// * If we have an existing authentication that is going to expire,
-			// never back off more than half of the amount of time remaining
-			// until expiration
-			// * Never back off more than 30 seconds multiplied by a random
-			// value between 1 and 2
-			// * Use randomness so that many clients won't keep hitting Vault
-			// at the same time
+			metrics.IncrCounter([]string{"nomad", "vault", "renew_failed"}, 1)
+			v.logger.Warn("got error or bad auth, so backing off", "error", err, "recoverable", recoverable)
 
-			// Set base values and add some backoff
-
-			v.logger.Printf("[WARN] vault: got error or bad auth, so backing off: %v", err)
-			switch {
-			case backoff < 5:
-				backoff = 5
-			case backoff >= 24:
-				backoff = 30
-			default:
-				backoff = backoff * 1.25
+			if !recoverable {
+				return
 			}
 
-			// Add randomness
-			backoff = backoff * (1.0 + rand.Float64())
-
-			maxBackoff := currentExpiration.Sub(time.Now()) / 2
-			if maxBackoff < 0 {
+			backoff = nextBackoff(backoff, currentExpiration)
+			if backoff < 0 {
 				// We have failed to renew the token past its expiration. Stop
 				// renewing with Vault.
-				v.logger.Printf("[ERR] vault: failed to renew Vault token before lease expiration. Shutting down Vault client")
+				v.logger.Error("failed to renew Vault token before lease expiration. Shutting down Vault client",
+					"error", err)
 				v.l.Lock()
 				v.connEstablished = false
 				v.connEstablishedErr = err
 				v.l.Unlock()
 				return
-
-			} else if backoff > maxBackoff.Seconds() {
-				backoff = maxBackoff.Seconds()
 			}
 
 			durationUntilRetry := time.Duration(backoff) * time.Second
-			v.logger.Printf("[INFO] vault: backing off for %v", durationUntilRetry)
+			v.logger.Info("backing off renewal", "retry", durationUntilRetry)
 
 			authRenewTimer.Reset(durationUntilRetry)
 		}
 	}
 }
 
+// nextBackoff returns the delay for the next auto renew interval, in seconds.
+// Returns negative value if past expiration
+//
+// It should increase the amount of backoff each time, with the following rules:
+//
+// * If token expired already despite earlier renewal attempts,
+//   back off for 1 minute + jitter
+// * If we have an existing authentication that is going to expire,
+// never back off more than half of the amount of time remaining
+// until expiration (with 5s floor)
+// * Never back off more than 30 seconds multiplied by a random
+// value between 1 and 2
+// * Use randomness so that many clients won't keep hitting Vault
+// at the same time
+func nextBackoff(backoff float64, expiry time.Time) float64 {
+	maxBackoff := time.Until(expiry) / 2
+
+	if maxBackoff < 0 {
+		// expiry passed
+		return 60 * (1.0 + rand.Float64())
+	}
+
+	switch {
+	case backoff >= 24:
+		backoff = 30
+	default:
+		backoff = backoff * 1.25
+	}
+
+	// Add randomness
+	backoff = backoff * (1.0 + rand.Float64())
+
+	if backoff > maxBackoff.Seconds() {
+		backoff = maxBackoff.Seconds()
+	}
+
+	if backoff < 5 {
+		backoff = 5
+	}
+
+	return backoff
+}
+
 // renew attempts to renew our Vault token. If the renewal fails, an error is
-// returned. This method updates the lastRenewed time
-func (v *vaultClient) renew() error {
+// returned.  The boolean indicates whether it's safe to attempt to renew again.
+// This method updates the currentExpiration time
+func (v *vaultClient) renew() (bool, error) {
+	// Track how long the request takes
+	defer metrics.MeasureSince([]string{"nomad", "vault", "renew"}, time.Now())
+
 	// Attempt to renew the token
 	secret, err := v.auth.RenewSelf(v.tokenData.CreationTTL)
 	if err != nil {
-		return err
+		// Check if there is a permission denied
+		recoverable := !structs.VaultUnrecoverableError.MatchString(err.Error())
+		return recoverable, fmt.Errorf("failed to renew the vault token: %v", err)
 	}
+
 	if secret == nil {
 		// It's possible for RenewSelf to return (nil, nil) if the
 		// response body from Vault is empty.
-		return fmt.Errorf("renewal failed: empty response from vault")
+		return true, fmt.Errorf("renewal failed: empty response from vault")
 	}
 
+	// these treated as transient errors, where can keep renewing
 	auth := secret.Auth
 	if auth == nil {
-		return fmt.Errorf("renewal successful but not auth information returned")
+		return true, fmt.Errorf("renewal successful but not auth information returned")
 	} else if auth.LeaseDuration == 0 {
-		return fmt.Errorf("renewal successful but no lease duration returned")
+		return true, fmt.Errorf("renewal successful but no lease duration returned")
 	}
 
-	v.lastRenewed = time.Now()
-	v.logger.Printf("[DEBUG] vault: successfully renewed server token")
-	return nil
+	v.extendExpiration(auth.LeaseDuration)
+
+	v.logger.Debug("successfully renewed server token")
+	return true, nil
 }
 
 // getWrappingFn returns an appropriate wrapping function for Nomad Servers
@@ -599,7 +706,6 @@ func (v *vaultClient) parseSelfToken() error {
 	if err := mapstructure.WeakDecode(secret.Data, &data); err != nil {
 		return fmt.Errorf("failed to parse Vault token's data block: %v", err)
 	}
-
 	root := false
 	for _, p := range data.Policies {
 		if p == "root" {
@@ -607,10 +713,9 @@ func (v *vaultClient) parseSelfToken() error {
 			break
 		}
 	}
-
-	// Store the token data
 	data.Root = root
 	v.tokenData = &data
+	v.extendExpiration(data.TTL)
 
 	// The criteria that must be met for the token to be valid are as follows:
 	// 1) If token is non-root or is but has a creation ttl
@@ -629,7 +734,7 @@ func (v *vaultClient) parseSelfToken() error {
 
 	var mErr multierror.Error
 	role := v.getRole()
-	if !root {
+	if !data.Root {
 		// All non-root tokens must be renewable
 		if !data.Renewable {
 			multierror.Append(&mErr, fmt.Errorf("Vault token is not renewable or root"))
@@ -661,7 +766,7 @@ func (v *vaultClient) parseSelfToken() error {
 	}
 
 	// Check we have the correct capabilities
-	if err := v.validateCapabilities(role, root); err != nil {
+	if err := v.validateCapabilities(role, data.Root); err != nil {
 		multierror.Append(&mErr, err)
 	}
 
@@ -718,12 +823,12 @@ func (v *vaultClient) validateCapabilities(role string, root bool) error {
 		if structs.VaultUnrecoverableError.MatchString(err.Error()) {
 			// Since we can't read permissions, we just log a warning that we
 			// can't tell if the Vault token will work
-			msg := fmt.Sprintf("Can not lookup token capabilities. "+
+			msg := fmt.Sprintf("can not lookup token capabilities. "+
 				"As such certain operations may fail in the future. "+
 				"Please give Nomad a Vault token with one of the following "+
 				"capabilities %q on %q so that the required capabilities can be verified",
 				vaultCapabilitiesCapability, vaultCapabilitiesLookupPath)
-			v.logger.Printf("[WARN] vault: %s", msg)
+			v.logger.Warn(msg)
 			return nil
 		} else {
 			multierror.Append(&mErr, err)
@@ -801,10 +906,12 @@ func (v *vaultClient) validateRole(role string) error {
 
 	// Read and parse the fields
 	var data struct {
-		ExplicitMaxTtl int `mapstructure:"explicit_max_ttl"`
-		Orphan         bool
-		Period         int
-		Renewable      bool
+		ExplicitMaxTtl      int `mapstructure:"explicit_max_ttl"`
+		TokenExplicitMaxTtl int `mapstructure:"token_explicit_max_ttl"`
+		Orphan              bool
+		Period              int
+		TokenPeriod         int `mapstructure:"token_period"`
+		Renewable           bool
 	}
 	if err := mapstructure.WeakDecode(rsecret.Data, &data); err != nil {
 		return fmt.Errorf("failed to parse Vault role's data block: %v", err)
@@ -816,11 +923,11 @@ func (v *vaultClient) validateRole(role string) error {
 		multierror.Append(&mErr, fmt.Errorf("Role must allow tokens to be renewed"))
 	}
 
-	if data.ExplicitMaxTtl != 0 {
+	if data.ExplicitMaxTtl != 0 || data.TokenExplicitMaxTtl != 0 {
 		multierror.Append(&mErr, fmt.Errorf("Role can not use an explicit max ttl. Token must be periodic."))
 	}
 
-	if data.Period == 0 {
+	if data.Period == 0 && data.TokenPeriod == 0 {
 		multierror.Append(&mErr, fmt.Errorf("Role must have a non-zero period to make tokens periodic."))
 	}
 
@@ -857,7 +964,6 @@ func (v *vaultClient) CreateToken(ctx context.Context, a *structs.Allocation, ta
 	if !v.Active() {
 		return nil, structs.NewRecoverableError(fmt.Errorf("Vault client not active"), true)
 	}
-
 	// Check if we have established a connection with Vault
 	if established, err := v.ConnectionEstablished(); !established && err == nil {
 		return nil, structs.NewRecoverableError(fmt.Errorf("Connection to Vault has not been established"), true)
@@ -882,6 +988,12 @@ func (v *vaultClient) CreateToken(ctx context.Context, a *structs.Allocation, ta
 		return nil, fmt.Errorf("Task does not require Vault policies")
 	}
 
+	// Set namespace for task
+	namespaceForTask := v.config.Namespace
+	if taskVault.Namespace != "" {
+		namespaceForTask = taskVault.Namespace
+	}
+
 	// Build the creation request
 	req := &vapi.TokenCreateRequest{
 		Policies: taskVault.Policies,
@@ -889,6 +1001,7 @@ func (v *vaultClient) CreateToken(ctx context.Context, a *structs.Allocation, ta
 			"AllocationID": a.ID,
 			"Task":         task,
 			"NodeID":       a.NodeID,
+			"Namespace":    namespaceForTask,
 		},
 		TTL:         v.childTTL,
 		DisplayName: fmt.Sprintf("%s-%s", a.ID, task),
@@ -904,16 +1017,24 @@ func (v *vaultClient) CreateToken(ctx context.Context, a *structs.Allocation, ta
 	var secret *vapi.Secret
 	var err error
 	role := v.getRole()
+
+	// Fetch client for task
+	taskClient, err := v.entHandler.clientForTask(v, namespaceForTask)
+	if err != nil {
+		return nil, err
+	}
+
 	if v.tokenData.Root && role == "" {
 		req.Period = v.childTTL
-		secret, err = v.auth.Create(req)
+		secret, err = taskClient.Auth().Token().Create(req)
 	} else {
 		// Make the token using the role
-		secret, err = v.auth.CreateWithRole(req, v.getRole())
+		secret, err = taskClient.Auth().Token().CreateWithRole(req, v.getRole())
 	}
 
 	// Determine whether it is unrecoverable
 	if err != nil {
+		err = fmt.Errorf("failed to create an alloc vault token: %v", err)
 		if structs.VaultUnrecoverableError.MatchString(err.Error()) {
 			return secret, err
 		}
@@ -932,7 +1053,7 @@ func (v *vaultClient) CreateToken(ctx context.Context, a *structs.Allocation, ta
 		validationErr = fmt.Errorf("Vault returned WrapInfo without WrappedAccessor. Secret warnings: %v", secret.Warnings)
 	}
 	if validationErr != nil {
-		v.logger.Printf("[WARN] vault: failed to CreateToken: %v", err)
+		v.logger.Warn("failed to CreateToken", "error", validationErr)
 		return nil, structs.NewRecoverableError(validationErr, true)
 	}
 
@@ -972,15 +1093,23 @@ func (v *vaultClient) LookupToken(ctx context.Context, token string) (*vapi.Secr
 
 // PoliciesFrom parses the set of policies returned by a token lookup.
 func PoliciesFrom(s *vapi.Secret) ([]string, error) {
+	return s.TokenPolicies()
+}
+
+// PolicyDataFrom parses the Data returned by a token lookup.
+// It should not be used to parse TokenPolicies as the list will not be
+// exhaustive.
+func PolicyDataFrom(s *vapi.Secret) (tokenData, error) {
 	if s == nil {
-		return nil, fmt.Errorf("cannot parse nil Vault secret")
+		return tokenData{}, fmt.Errorf("cannot parse nil Vault secret")
 	}
 	var data tokenData
+
 	if err := mapstructure.WeakDecode(s.Data, &data); err != nil {
-		return nil, fmt.Errorf("failed to parse Vault token's data block: %v", err)
+		return tokenData{}, fmt.Errorf("failed to parse Vault token's data block: %v", err)
 	}
 
-	return data.Policies, nil
+	return data, nil
 }
 
 // RevokeTokens revokes the passed set of accessors. If committed is set, the
@@ -1024,7 +1153,7 @@ func (v *vaultClient) RevokeTokens(ctx context.Context, accessors []*structs.Vau
 			return nil
 		}
 
-		v.logger.Printf("[WARN] vault: failed to revoke tokens. Will reattempt until TTL: %v", err)
+		v.logger.Warn("failed to revoke tokens. Will reattempt until TTL", "error", err)
 		v.storeForRevocation(accessors)
 		return nil
 	} else if !committed {
@@ -1034,7 +1163,7 @@ func (v *vaultClient) RevokeTokens(ctx context.Context, accessors []*structs.Vau
 	}
 
 	if err := v.purgeFn(accessors); err != nil {
-		v.logger.Printf("[ERR] vault: failed to purge Vault accessors: %v", err)
+		v.logger.Error("failed to purge Vault accessors", "error", err)
 		v.storeForRevocation(accessors)
 		return nil
 	}
@@ -1045,18 +1174,31 @@ func (v *vaultClient) RevokeTokens(ctx context.Context, accessors []*structs.Vau
 	return nil
 }
 
+func (v *vaultClient) MarkForRevocation(accessors []*structs.VaultAccessor) error {
+	if !v.Enabled() {
+		return nil
+	}
+
+	if !v.Active() {
+		return fmt.Errorf("Vault client not active")
+	}
+
+	v.storeForRevocation(accessors)
+	return nil
+}
+
 // storeForRevocation stores the passed set of accessors for revocation. It
 // captures their effective TTL by storing their create TTL plus the current
 // time.
 func (v *vaultClient) storeForRevocation(accessors []*structs.VaultAccessor) {
 	v.revLock.Lock()
-	v.statsLock.Lock()
+
 	now := time.Now()
 	for _, a := range accessors {
-		v.revoking[a] = now.Add(time.Duration(a.CreationTTL) * time.Second)
+		if _, ok := v.revoking[a]; !ok {
+			v.revoking[a] = now.Add(time.Duration(a.CreationTTL) * time.Second)
+		}
 	}
-	v.stats.TrackedForRevoke = len(v.revoking)
-	v.statsLock.Unlock()
 	v.revLock.Unlock()
 }
 
@@ -1085,7 +1227,7 @@ func (v *vaultClient) parallelRevoke(ctx context.Context, accessors []*structs.V
 		handlers = maxParallelRevokes
 	}
 
-	// Create the Vault Tokens
+	// Revoke the Vault Token Accessors
 	input := make(chan *structs.VaultAccessor, handlers)
 	for i := 0; i < handlers; i++ {
 		g.Go(func() error {
@@ -1096,7 +1238,8 @@ func (v *vaultClient) parallelRevoke(ctx context.Context, accessors []*structs.V
 						return nil
 					}
 
-					if err := v.auth.RevokeAccessor(va.Accessor); err != nil {
+					err := v.auth.RevokeAccessor(va.Accessor)
+					if err != nil && !strings.Contains(err.Error(), "invalid accessor") {
 						return fmt.Errorf("failed to revoke token (alloc: %q, node: %q, task: %q): %v", va.AllocID, va.NodeID, va.Task, err)
 					}
 				case <-pCtx.Done():
@@ -1123,10 +1266,23 @@ func (v *vaultClient) parallelRevoke(ctx context.Context, accessors []*structs.V
 	return g.Wait()
 }
 
+// maxVaultRevokeBatchSize is the maximum tokens a revokeDaemon should revoke
+// and purge at any given time.
+//
+// Limiting the revocation batch size is beneficial for few reasons:
+// * A single revocation failure of any entry in batch result into retrying the whole batch;
+//   the larger the batch is the higher likelihood of such failure
+// * Smaller batch sizes result into more co-operativeness: provides hooks for
+//   reconsidering token TTL and leadership steps down.
+// * Batches limit the size of the Raft message purging tokens. Due to bugs
+//   pre-0.11.3, expired tokens were not properly purged, so users upgrading from
+//   older versions may have huge numbers (millions) of expired tokens to purge.
+const maxVaultRevokeBatchSize = 1000
+
 // revokeDaemon should be called in a goroutine and is used to periodically
 // revoke Vault accessors that failed the original revocation
 func (v *vaultClient) revokeDaemon() {
-	ticker := time.NewTicker(vaultRevocationIntv)
+	ticker := time.NewTicker(v.revocationIntv)
 	defer ticker.Stop()
 
 	for {
@@ -1134,7 +1290,7 @@ func (v *vaultClient) revokeDaemon() {
 		case <-v.tomb.Dying():
 			return
 		case now := <-ticker.C:
-			if established, _ := v.ConnectionEstablished(); !established {
+			if established, err := v.ConnectionEstablished(); !established || err != nil {
 				continue
 			}
 
@@ -1146,18 +1302,33 @@ func (v *vaultClient) revokeDaemon() {
 				continue
 			}
 
-			// Build the list of allocations that need to revoked while pruning any TTL'd checks
-			revoking := make([]*structs.VaultAccessor, 0, len(v.revoking))
+			// Build the list of accessors that need to be revoked while pruning any TTL'd checks
+			toRevoke := len(v.revoking)
+			if toRevoke > v.maxRevokeBatchSize {
+				v.logger.Info("batching tokens to be revoked",
+					"to_revoke", toRevoke, "batch_size", v.maxRevokeBatchSize,
+					"batch_interval", v.revocationIntv)
+				toRevoke = v.maxRevokeBatchSize
+			}
+			revoking := make([]*structs.VaultAccessor, 0, toRevoke)
+			ttlExpired := []*structs.VaultAccessor{}
 			for va, ttl := range v.revoking {
 				if now.After(ttl) {
-					delete(v.revoking, va)
+					ttlExpired = append(ttlExpired, va)
 				} else {
 					revoking = append(revoking, va)
 				}
+
+				// Batches should consider tokens to be revoked
+				// as well as expired tokens to ensure the Raft
+				// message is reasonably sized.
+				if len(revoking)+len(ttlExpired) >= toRevoke {
+					break
+				}
 			}
 
 			if err := v.parallelRevoke(context.Background(), revoking); err != nil {
-				v.logger.Printf("[WARN] vault: background token revocation errored: %v", err)
+				v.logger.Warn("background token revocation errored", "error", err)
 				v.revLock.Unlock()
 				continue
 			}
@@ -1165,10 +1336,14 @@ func (v *vaultClient) revokeDaemon() {
 			// Unlock before a potentially expensive operation
 			v.revLock.Unlock()
 
+			// purge all explicitly revoked as well as ttl expired tokens
+			// and only remove them locally on purge success
+			revoking = append(revoking, ttlExpired...)
+
 			// Call the passed in token revocation function
 			if err := v.purgeFn(revoking); err != nil {
 				// Can continue since revocation is idempotent
-				v.logger.Printf("[ERR] vault: token revocation errored: %v", err)
+				v.logger.Error("token revocation errored", "error", err)
 				continue
 			}
 
@@ -1177,12 +1352,9 @@ func (v *vaultClient) revokeDaemon() {
 
 			// Can delete from the tracked list now that we have purged
 			v.revLock.Lock()
-			v.statsLock.Lock()
 			for _, va := range revoking {
 				delete(v.revoking, va)
 			}
-			v.stats.TrackedForRevoke = len(v.revoking)
-			v.statsLock.Unlock()
 			v.revLock.Unlock()
 
 		}
@@ -1214,29 +1386,66 @@ func (v *vaultClient) setLimit(l rate.Limit) {
 	v.limiter = rate.NewLimiter(l, int(l))
 }
 
-// Stats is used to query the state of the blocked eval tracker.
-func (v *vaultClient) Stats() *VaultStats {
+func (v *vaultClient) Stats() map[string]string {
+	stat := v.stats()
+
+	expireTimeStr := ""
+
+	if !stat.TokenExpiry.IsZero() {
+		expireTimeStr = stat.TokenExpiry.Format(time.RFC3339)
+	}
+
+	return map[string]string{
+		"tracked_for_revoked": strconv.Itoa(stat.TrackedForRevoke),
+		"token_ttl":           stat.TokenTTL.Round(time.Second).String(),
+		"token_expire_time":   expireTimeStr,
+	}
+}
+
+func (v *vaultClient) stats() *VaultStats {
 	// Allocate a new stats struct
 	stats := new(VaultStats)
 
-	v.statsLock.RLock()
-	defer v.statsLock.RUnlock()
+	v.revLock.Lock()
+	stats.TrackedForRevoke = len(v.revoking)
+	v.revLock.Unlock()
 
-	// Copy all the stats
-	stats.TrackedForRevoke = v.stats.TrackedForRevoke
+	v.currentExpirationLock.Lock()
+	stats.TokenExpiry = v.currentExpiration
+	v.currentExpirationLock.Unlock()
+
+	if !stats.TokenExpiry.IsZero() {
+		stats.TokenTTL = time.Until(stats.TokenExpiry)
+	}
 
 	return stats
 }
 
 // EmitStats is used to export metrics about the blocked eval tracker while enabled
-func (v *vaultClient) EmitStats(period time.Duration, stopCh chan struct{}) {
+func (v *vaultClient) EmitStats(period time.Duration, stopCh <-chan struct{}) {
 	for {
 		select {
 		case <-time.After(period):
-			stats := v.Stats()
+			stats := v.stats()
 			metrics.SetGauge([]string{"nomad", "vault", "distributed_tokens_revoking"}, float32(stats.TrackedForRevoke))
+			metrics.SetGauge([]string{"nomad", "vault", "token_ttl"}, float32(stats.TokenTTL/time.Millisecond))
+
 		case <-stopCh:
 			return
 		}
 	}
 }
+
+// extendExpiration sets the current auth token expiration record to ttLSeconds seconds from now
+func (v *vaultClient) extendExpiration(ttlSeconds int) {
+	v.currentExpirationLock.Lock()
+	v.currentExpiration = time.Now().Add(time.Duration(ttlSeconds) * time.Second)
+	v.currentExpirationLock.Unlock()
+}
+
+// VaultVaultNoopDelegate returns the default vault api auth token handler
+type VaultNoopDelegate struct{}
+
+func (e *VaultNoopDelegate) clientForTask(v *vaultClient, namespace string) (*vapi.Client, error) {
+	return v.client, nil
+}
